#![forbid(elided_lifetimes_in_paths, unsafe_code)]

use aoc23::read;
use chumsky::prelude::*;
use std::collections::{HashMap, HashSet, VecDeque};

enum Instruction {
	Left,
	Right
}

impl Instruction {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		choice((
			just('L').map(|_| Self::Left),
			just('R').map(|_| Self::Right)
		))
	}
}

#[derive(Debug)]
struct Node {
	left: String,
	right: String
}

fn parse_node_name() -> impl Parser<char, String, Error = Simple<char>> {
	none_of([',', ')', ' '])
		.repeated()
		.at_least(1)
		.map(|chars| chars.into_iter().collect())
}

impl Node {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		just("(")
			.ignore_then(parse_node_name())
			.then_ignore(just(", "))
			.then(parse_node_name())
			.then_ignore(just(")"))
			.map(|(left, right)| Self { left, right })
	}
}

#[derive(Debug)]
struct Network {
	nodes: HashMap<String, Node>
}

impl Network {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		parse_node_name()
			.then_ignore(just(" = "))
			.then(Node::parser())
			.then_ignore(just("\n"))
			.repeated()
			.map(|nodes| Self {
				nodes: nodes.into_iter().collect()
			})
	}

	fn node(&self, node: &str) -> &Node {
		self.nodes
			.get(node)
			.unwrap_or_else(|| panic!("Failed to get node {node:?}"))
	}

	fn left(&self, node: &str) -> &str {
		&self.node(node).left
	}

	fn right(&self, node: &str) -> &str {
		&self.node(node).right
	}
}

fn parser() -> impl Parser<char, (Vec<Instruction>, Network), Error = Simple<char>> {
	Instruction::parser()
		.repeated()
		.at_least(1)
		.then_ignore(just("\n\n"))
		.then(Network::parser())
		.then_ignore(end())
}

#[derive(Copy, Clone, Debug)]
struct Ghost<'a> {
	node: &'a str,
	i: usize,
	steps: u128
}

impl<'a> Ghost<'a> {
	fn new(node: &'a str) -> Self {
		Self {
			node,
			i: 0,
			steps: 0
		}
	}

	fn advance(&mut self, ghost: Ghost<'a>) {
		self.node = ghost.node;
		self.i = ghost.i;
		self.steps += ghost.steps;
	}
}

fn main() -> anyhow::Result<()> {
	let (instructions, network) = read("inputs/day8.txt", parser())?;

	// build a lookup map of left/right predecessors
	let mut left_pred: HashMap<&str, Vec<&str>> = HashMap::new();
	let mut right_pred: HashMap<&str, Vec<&str>> = HashMap::new();
	for (node, next) in &network.nodes {
		left_pred.entry(&next.left).or_default().push(node);
		right_pred.entry(&next.right).or_default().push(node);
	}
	eprintln!("Built pred lookup map");

	// build a lookup map from instruction position and node with the length to the next
	// finish node
	let mut dp: HashMap<(&str, usize), Ghost<'_>> = HashMap::new();
	let mut q: VecDeque<(&str, usize)> = VecDeque::new();
	for node in network
		.nodes
		.keys()
		.map(|node| node.as_str())
		.filter(|node| node.ends_with('Z'))
	{
		for i in 0 .. instructions.len() {
			dp.insert((node, i), Ghost { node, i, steps: 0 });
			q.push_back((node, i));
		}
	}
	while let Some((node, i)) = q.pop_front() {
		let ghost = dp[&(node, i)];
		let i_pred = match i {
			0 => instructions.len() - 1,
			i => i - 1
		};
		let pred = match instructions[i_pred] {
			Instruction::Left => left_pred.get(node),
			Instruction::Right => right_pred.get(node)
		};
		for node_pred in pred.into_iter().flatten() {
			let new = (*node_pred, i_pred);
			#[allow(clippy::map_entry)] // lint is opinionated garbage
			if !dp.contains_key(&new) {
				dp.insert(new, Ghost {
					node: ghost.node,
					i: ghost.i,
					steps: ghost.steps + 1
				});
				q.push_back(new);
				if dp.len() % 1_000_000 == 0 {
					dbg!(dp.len());
				}
			}
		}
		if ghost.steps == 0 {
			dp.remove(&(node, i));
		}
	}
	eprintln!("Built ghost lookup map");
	// for (key, value) in &dp {
	// eprintln!("\t{key:?}\t = \t{value:?}");
	// }

	// let mut curr_node = "AAA";
	// let mut i = 0;
	// let mut steps = 0_u128;
	// while curr_node != "ZZZ" {
	// 	curr_node = match instructions[i] {
	// 		Instruction::Left => network.left(curr_node),
	// 		Instruction::Right => network.right(curr_node)
	// 	};
	// 	i = (i + 1) % instructions.len();
	// 	steps += 1;
	// }
	// println!("{steps}");

	let mut ghosts: Vec<Ghost<'_>> = network
		.nodes
		.keys()
		.filter(|node| node.ends_with('A'))
		.map(|node| Ghost::new(node))
		.collect();
	let mut progress = 0;
	loop {
		if ghosts[0].steps > 0
			&& ghosts
				.iter()
				.skip(1)
				.all(|ghost| ghost.steps == ghosts[0].steps)
		{
			println!("{}", ghosts[0].steps);
			break;
		}

		let (ghost_index, ghost) = ghosts
			.iter()
			.enumerate()
			.min_by_key(|(_, ghost)| ghost.steps)
			.unwrap();
		if ghost.steps - progress >= 1_000_000_000_000 {
			progress = ghost.steps;
			eprintln!("progress: {progress}");
		}

		// eprint!("{ghost:?}\t -> \t");
		let ghost = dp[&(ghost.node, ghost.i)];
		// eprintln!("{ghost:?}");
		ghosts[ghost_index].advance(ghost);
	}

	Ok(())
}
